#include "inifile.cpp"


// read in parameters from the .ini file, using a c++ class declared in inifile.h
void read_parameters(const char* filename) {
  inifile ini(filename);

  // syntax: to read e.g. a double value: ini.getdbl("keyword", default_value, set-number);
  // read integer value: ini.getlng(...);

  dt = ini.getdbl("dt", 0.05, 0);
  mpcsteps = ini.getlng("mpcsteps", 10000, 0);
  mdsteps = ini.getlng("mdsteps", 20, 0);
  rho = ini.getdbl("rho", 10.0, 0);
  Lx = ini.getlng("Lx", 50, 0);
  Ly = ini.getlng("Ly", 50, 0);
  alpha = ini.getdbl("alpha", M_PI/2.0, 0);
  temperature = ini.getdbl("temperature", 1.0, 0);
  grav = ini.getdbl("grav", 0.004, 0);
  obsMass = ini.getdbl("obsMass", 10.0, 0);
  spring_force = ini.getdbl("spring_force", 1000.0, 0);
  radius = ini.getlng("radius", 5, 0);
  vis_cellsize = ini.getlng("vis_cellsize", 2, 0);
  measurement_interval = ini.getlng("measurement_interval", 100, 0);
  start_flow_measurement = ini.getlng("start_flow_measurement", 5000, 0);
  fLx = int(ceil(Lx/(double)(vis_cellsize))+0.5);        /* number of cells in the flow field visualization */
  fLy = int(ceil(Ly/(double)(vis_cellsize))+0.5);
  dLx = (double)Lx;
  dLy = (double)Ly;

}



// Initialize variables, allocate memory, define starting positions and velocities. Nothing too interesting here.
void initialize()
{
  int i;

  init_genrand(17235321);   /* initializing random number generators */

  Ncell=Lx*(Ly+1);
  N=(int) ((dLx)*(dLy)*rho);    /* number of fluid particles */
  flowstorecount = 0;

  obsStartx = (double *) malloc(sizeof(double)*(2*radius+1)*(2*radius+1));  /* allocate memory */
  obsStarty = (double *) malloc(sizeof(double)*(2*radius+1)*(2*radius+1));
  place_obstacle();
  head=(int *) malloc(sizeof(int)*Ncell);       
  list=(int *) malloc(sizeof(int)*(N+Nobs));
  rx=(double *) malloc(sizeof(double)*(N+Nobs));
  ry=(double *) malloc(sizeof(double)*(N+Nobs));
  vx=(double *) malloc(sizeof(double)*(N+Nobs));
  vy=(double *) malloc(sizeof(double)*(N+Nobs));
  flowfieldx = (double *) malloc(sizeof(double)*fLx*fLy); // fLx: number of vectors in the flow-field visualization in x-direction
  flowfieldy = (double *) malloc(sizeof(double)*fLx*fLy);
  flowcellmass = (unsigned int *) malloc(sizeof(unsigned int)*fLx*fLy);
  for (i = 0; i < fLx*fLy; i++) {
    flowfieldx[i] = 0.0;
    flowfieldy[i] = 0.0;
    flowcellmass[i] = 0;
  }

  initialPositions();
  initialVelocities();
}

// free the allocated memory after program completion
void cleanup()
{
  free(list);
  free(head);
  free(rx);
  free(ry);
  free(vx);
  free(vy);
  free(flowfieldx);
  free(flowfieldy);
  free(flowcellmass);
  free(obsStartx);
  free(obsStarty);
}

double gauss()
{
  /* converts uniformly distributed random*/
  /* numbers to gaussian distributed ones*/
  /* (Numerical Recipes in C, Box-Mueller method) */

  static int iset=0;
  static double gset;
  double fac,rsq,v1,v2;

  if(iset==0)
    {
      do
        {
          v1=2.0*UNI-1.0;
          v2=2.0*UNI-1.0;
          rsq=v1*v1+v2*v2;
        }while(rsq>=1.0||rsq==0.0);
      fac=sqrt(-2.0*log(rsq)/rsq);
      gset=v1*fac;
      iset=1;
      return v2*fac;
    }
  else
    {
      iset=0;
      return gset;
    }
}
